"""
    NUMPY CONTENT DECODERS

    Handle Decoding of Content in to appropriate format from request as numpy array.

    e.g. 
        list of lists -> np.array.
"""
import csv
import io
from PIL import Image
import numpy as np

def image_to_numpy(image_bytes, convert: str = None):
    """Convert an image bytes string into numpy through PIL.
    Args:
        bytes_like (str): bytes serialized image string.
        convert (str): pillow image convert mode
    Returns:
        (np.array): Converted numpy array.
    """
    pillow_image =  Image.open(
        io.BytesIO(image_bytes)
    )
    if convert is not None:
        pillow_image = pillow_image.convert(convert)
    return np.asarray(pillow_image)



"""Predict Request API utilties"""
import io
import json
import requests
import json

def execute_request(url, headers, data, **kwargs):
    """compiles and executes request to API"""
    return requests.post(
        url=url,
        headers=headers,
        data=data
    )

def send_json(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = json.load(file_)

    response = execute_request(
        url=host,
        headers={'Content-Type': 'application/json'},
        data=json.dumps(data)
    )
    if response.status_code == 200:
        return "{}".format(json.dumps(response.json(), indent=4, separators=(',', ': '), sort_keys=True))
    else:
        raise requests.exceptions.RequestException("error ({}): {}".format(response.status_code, response.raise_for_status()))

def send_csv(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = file_.read()

    response = execute_request(
        url=host,
        headers={'Content-Type': 'text/csv'},
        data=data
    )

    if response.status_code == 200:
        return response.text
    else:
        raise requests.exceptions.RequestException("error ({}): {}".format(response.status_code, response.raise_for_status()))

def send_image_jpeg(filepath, host):
    """Send a image jpeg payload as bytes in request providing a filepath."""

    with open(filepath, 'rb') as file_:
        byte_content = io.BytesIO(file_.read())

    response = execute_request(
        url=host,
        headers={'Content-Type': 'image/jpeg'},
        data=byte_content
    )

    if response.status_code == 200:
        return "{}".format(json.dumps(response.json(), indent=4, separators=(',', ': '), sort_keys=True))
    else:
        raise requests.exceptions.RequestException("error ({}): {}".format(response.status_code, response.raise_for_status()))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
#
# Disclaimer: This code can be found here: https://github.com/aws/sagemaker-training-toolkit/blob/master/test/unit/test_encoder.py
#
import io
import pytest
import numpy as np
from PIL import Image

from mldock.platform_helpers.mldock.inference.content_decoders.image import numpy as numpy_image_decoders

@pytest.fixture
def image_bytes():
    """reads image as bytes string"""
    with open('tests/api/fixtures/eight.png', 'rb') as file_:
        return file_.read()

@pytest.fixture
def image_array():
    """reads image as bytes string"""
    return np.asarray(Image.open('tests/api/fixtures/eight.png'))

class TestNumpyImageDecoders:
    """Tests the numpy decoder methods"""

    @staticmethod
    def test_image_to_numpy(image_bytes, image_array):
        """test csv data is correctly decoded to numpy array"""
        actual = numpy_image_decoders.image_to_numpy(image_bytes)

        np.testing.assert_equal(actual, image_array)



"""Test Container commands"""
import os
from pathlib import Path
from mock import patch
from click.testing import CliRunner
import pytest
import logging
import tempfile
from mldock.__main__ import cli
from mldock.platform_helpers import utils

logger=logging.getLogger('mldock')

@pytest.fixture
def mldock_config_aws():

    return {
        'container_dir': 'container',
        'image_name': 'my_app',
        'mldock_module_dir': 'src',
        'requirements_dir': 'requirements.txt',
        'template': 'generic'
    }

class TestContainerCommands:

    @staticmethod
    def test_command_init_w_no_prompt_runs_successfully():
        """
            Tests that the init no-prompt command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmp_dir:

            mldock_project = Path(tmp_dir, 'my_app')

            utils._copy_boilerplate_to_dst(
                src='tests/command/fixtures/base_container',
                dst=mldock_project,
                remove_first=True
            )

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(cli=cli, args=['container', 'init', '--dir', mldock_project, '--no-prompt'])

        assert result.exit_code == 0, result.output

    @staticmethod
    def test_command_init_w_user_input_runs_successfully():
        """
            Tests that the init with user input command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir:

            mldock_project = Path(tmp_dir, 'my_app')

            utils._copy_boilerplate_to_dst(
                src='tests/command/fixtures/base_container',
                dst=mldock_project,
                remove_first=True
            )

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(
                cli=cli,
                args=['container', 'init', '--dir', mldock_project],
                input='my_app\ngeneric\nsrc\ncontainer\nrequirements.txt\n\n\n\n\n'
            )

        assert result.exit_code == 0, result.output

    @staticmethod
    def test_command_init_w_template_no_prompt_runs_successfully(mldock_config_aws):
        """
            Tests that the init with user input command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()
        # tmp_dir = 'my_app'
        with tempfile.TemporaryDirectory() as tmp_dir:

            # mldock can't write a new file in tmp directory
            # init an empty file
            mldock_project = Path(tmp_dir, 'my_app')
            mldock_filepath = Path(mldock_project,'mldock.json')

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(
                cli=cli,
                args=['container', 'init', '--dir', mldock_project, '--template', 'generic', '--no-prompt'],
                input='yes\n'
            )

            assert result.exit_code == 0, result.output

            assert mldock_filepath.is_file(), "Failed. No mldock.json was found"

            mldock_config = utils._read_json(mldock_filepath)
            mldock_config_aws.update({'image_name': 'my_app'})

        assert isinstance(mldock_config.pop('data'), list), "Failed. data was not a list"
        assert isinstance(mldock_config.pop('environment'), dict), "Failed. environment was not a dict"
        assert isinstance(mldock_config.pop('hyperparameters'), dict), "Failed. hyperparameters was not a dict"
        assert isinstance(mldock_config.pop('model'), list), "Failed. model was not a list"
        assert isinstance(mldock_config.pop('stages'), dict), "Failed. stages was not a dict"

        assert mldock_config == mldock_config_aws, "Failed. MLDOCK container project config was incorrect"

    @staticmethod
    def test_command_init_w_template_no_prompt_runs_fails_gracefully(mldock_config_aws):
        """
            Tests that the init with user input command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tmp_dir:

            # mldock can't write a new file in tmp directory
            # init an empty file
            mldock_project = Path(tmp_dir, 'my_app')
            mldock_filepath = Path(mldock_project,'mldock.json')

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(
                cli=cli,
                args=['container', 'init', '--dir', tmp_dir, '--template', 'generic', '--no-prompt'],
                input='\n'
            )

            assert result.exit_code == 1, result.output

            assert not mldock_filepath.is_file(), "Failed. A mldock.json was found"

    @staticmethod
    def test_command_init_w_template_no_prompt_w_param_runs_successfully(mldock_config_aws):
        """
            Tests that the init with user input command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmp_dir:

            mldock_project = Path(tmp_dir, 'my_app')
            mldock_filepath = Path(mldock_project,'mldock.json')

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(
                cli=cli,
                args=[
                    'container', 'init',
                    '--dir', mldock_project,
                    '--template', 'generic',
                    '--no-prompt',
                    '-p', 'my_param', 'value',
                    '-e', 'my_var', 'value'
                ],
                input='yes\n'
            )

            assert result.exit_code == 0, result.output

            assert mldock_filepath.is_file(), "Failed. No mldock.json was found"

            mldock_config = utils._read_json(mldock_filepath)

            mldock_config_aws.update({'image_name': 'my_app'})

        assert isinstance(mldock_config.pop('data'), list), "Failed. data was not a list"
        assert isinstance(mldock_config.pop('model'), list), "Failed. model was not a list"
        assert isinstance(mldock_config.pop('stages'), dict), "Failed. stages was not a dict"

        assert mldock_config.pop('environment') == {'my_var': 'value'}, "Failed. environment were not correct."
        assert mldock_config.pop('hyperparameters') == {'my_param': 'value'}, "Failed. hyperparameters were not correct."
        assert mldock_config == mldock_config_aws, "Failed. MLDOCK container project config was incorrect"

